from random import sample
from typing import Union

from numpy import abs as np_abs, ndarray, empty
from numpy import copy, int8
from numpy import sum as np_sum
from numpy.random import RandomState
import numpy as np

from .simulation import Simulation

BIT_TYPE = int8


# TODO add documentation


def random_inputs(n: int, N: int, seed: int) -> ndarray:
    return 2 * RandomState(seed).randint(0, 2, (N, n), dtype=BIT_TYPE) - 1


def transform_challenge_01_to_11(challenge: ndarray) -> ndarray:  # TODO remove
    res = copy(challenge)
    res[res == 1] = -1
    res[res == 0] = 1
    return res


def transform_challenge_11_to_01(challenge: ndarray) -> ndarray:  # TODO remove
    res = copy(challenge)
    res[res == 1] = 0
    res[res == -1] = 1
    return res


def approx_stabilities(instance: Simulation, num: int, reps: int, seed: int) -> ndarray:
    challenges = random_inputs(instance.challenge_length, num, seed)
    responses = empty((reps, num), dtype=BIT_TYPE)
    for i in range(reps):
        responses[i, :] = instance.eval(challenges)
    return 0.5 + 0.5 * np_abs(np_sum(responses, axis=0)) / reps


class ChallengeInformationSet:
    # TODO fix type annotation when removing python 3.7 support

    def __init__(self, challenges: ndarray, information: ndarray) -> None:
        if challenges.shape[0] != information.shape[0]:
            raise ValueError('Must supply an equal number of challenges and information about these challenges.')
        self.challenges = challenges
        self.information = information
        self.N = len(self.challenges)

    @property
    def challenge_length(self) -> int:
        return self.challenges.shape[1]

    def __len__(self) -> int:
        return self.challenges.shape[0]

    def __getitem__(self, item: Union[slice, int]) -> object:
        return self.__class__(self.challenges[item], self.information[item])

    def random_subset(self, N: Union[int, float]) -> object:
        if N < 1:
            N = int(self.N * N)
        return self[sample(range(self.N), N)]

    def block_subset(self, i: int, total: int) -> object:
        return self[int(i / total * self.N):int((i + 1) / total * self.N)]


class ChallengeResponseSet(ChallengeInformationSet):

    @classmethod
    def from_simulation(cls, instance: Simulation, N: int, seed: int, r: int = 1) -> ChallengeInformationSet:
        challenges = random_inputs(instance.challenge_length, N, seed)
        crp_set = cls(
            challenges=challenges,
            responses=instance.r_eval(r, challenges)
        )
        crp_set.instance = instance
        return crp_set

    def __init__(self, challenges: ndarray, responses: ndarray) -> None:
        super().__init__(challenges, responses)
        self.responses = responses

    @property
    def response_length(self) -> int:
        return self.responses.shape[1]


class ChallengeReliabilitySet(ChallengeInformationSet):

    @classmethod
    def from_simulation(cls, instance: Simulation, N: int, seed: int, r: int = 5) -> ChallengeInformationSet:
        # noinspection PyTypeChecker
        return cls.from_challenge_response_set(ChallengeResponseSet.from_simulation(instance, N, seed, r))

    @classmethod
    def from_challenge_response_set(cls, crp_set: ChallengeResponseSet) -> ChallengeInformationSet:
        return cls(
            challenges=crp_set.challenges,
            reliabilities=np.average(crp_set.responses, axis=-1),
        )

    def __init__(self, challenges: ndarray, reliabilities: ndarray) -> None:
        super().__init__(challenges, reliabilities)
        self.reliabilities = reliabilities



from itertools import combinations
from typing import List

import numpy as np

from ..io import random_inputs, ChallengeResponseSet
from ..simulation import Simulation


def reliability_data(responses: np.ndarray) -> np.ndarray:
    r"""
    Computes the reliability of response data of a PUF instance.

    Given responses of a PUF instance in an array of shape :math:`(N, m, r)`, where
    :math:`N` is the number of challenges queries,
    :math:`m` is the response length of the PUF,
    :math:`r` is the number of repeated queries for each challenge,
    this function computes the empirical reliability of the PUF responses, i.e. an approximation of

    .. math:: \Pr_\text{eval} \left[ \text{eval}(x) = \text{eval}(x) \right]

    where :math:`\text{eval}(x)` is the response of the (noisy) PUF when evaluated on input :math:`x`, and the
    probability is taken over the noise in the evaluation process.

    An array of shape :math:`(N,m)` is returned, estimating above probability for each challenge and each response bit.
    To obtain a the reliability for each response bit, average along the first axis, to obtain the general reliability,
    average over all axes.

    >>> from pypuf.metrics import reliability_data
    >>> from numpy import average, array
    >>> responses = array([[[1, 1, -1]], [[1, 1, 1]], [[1, 1, 1]], [[1, 1, 1]]])
    >>> average(reliability_data(responses), axis=0)
    array([0.83333333])
    """
    return np.absolute(np.average(responses, axis=-1))


def reliability(instance: Simulation, seed: int, N: int = 10000, r: int = 17) -> np.ndarray:
    r"""
    For a given simulation, estimates simulated reliability.

    Randomly generates :math:`N` challenges using the :math:`\mathtt{seed}`, then queries each challenge :math:`r`
    times, obtaining a response array of shape :math:`(N, m, r)`, where :math:`m` is the response length of the given
    simulation.

    Then applies :func:`reliability_data` to determine the reliability of each respones bit on each challenge.

    Returns a float-array of shape :math:`(N, m)`, giving the reliability of each response bit on each challenge.
    To obtain the general reliability for each response bit, average along the first axis. To obtain the total
    reliability of this instance, average across all axes. Note that bad reliability on single response bits may be
    problematic.

    >>> from pypuf.simulation import XORArbiterPUF
    >>> from pypuf.metrics import reliability
    >>> from numpy import average
    >>> puf = XORArbiterPUF(n=128, k=2, seed=1, noisiness=.2)
    >>> average(reliability(puf, seed=2), axis=0)
    array([0.77729412])
    """
    inputs = random_inputs(n=instance.challenge_length, N=N, seed=seed)
    return np.absolute(reliability_data(instance.r_eval(r, inputs)))


def uniqueness_data(responses: np.ndarray) -> np.ndarray:
    r"""
    Estimates the uniqueness of responses of a set of PUF instances per challenge.

    Given PUF responses in :math:`\{-1,1\}^m` in an array of shape :math:`(l, N, m)`, where
    :math:`l \in \mathbb{N}` is the number of PUF instances,
    :math:`N \in \mathbb{N}` is the number of challenges queried, and
    :math:`m \in \mathbb{N}` is the response length,
    returns an estimate of the uniqueness of PUFs on challenges in this set, i.e. an approximation of

    .. math::
        1 - 2 \cdot E_{f, g; f \neq g} \left[
            \left|
                \frac{1}{2} - \Pr_x \left[ f(x) = g(x) \right]
            \right|
        \right]

    where the challenges :math:`x` are from the (implicitly) provided challenges, :math:`f(x)` and :math:`g(x)` are
    the responses of PUFs :math:`f` and :math:`g` described in the provided data.

    For perfectly unique PUFs and for any :math:`f` and :math:`g`, we expect
    :math:`\Pr_x \left[ f(x) = g(x) \right] = 1/2`, hence the uniqueness to be 1.
    On the other hand, if any two :math:`f` and :math:`g` have similarity higher (or lower) than :math:`1/2`, then
    the uniqueness will be lower than 1.

    Returns an array of shape :math:`(m,)` with uniqueness values per response bit.

    >>> from numpy import random, array
    >>> from pypuf.metrics import uniqueness_data, similarity_data
    >>> prng = random.default_rng(seed=1)
    >>> # generate **different** random responses using numpy's pseudo-random generator
    >>> responses = array([2 * random.default_rng(seed).integers(0, 2, size=(1000, 3)) - 1 for seed in range(4)])
    >>> uniqueness_data(responses)
    array([0.97333333, 0.979     , 0.971     ])
    >>> # generate **same** random responses using numpy's pseudo-random generator
    >>> responses = array([2 * random.default_rng(1).integers(0, 2, size=(1000, 1)) - 1 for seed in range(4)])
    >>> uniqueness_data(responses)
    array([0.])
    """
    if len(responses.shape) == 2:
        # TODO fix LTFArray to return responses of shape (N, 1)
        l, N = responses.shape
        responses = responses.reshape((l, N, 1))

    l, _, m = responses.shape
    similarities = np.empty(shape=(l * (l - 1) // 2, m))
    for idx, (r1, r2) in enumerate(combinations(responses, 2)):
        similarities[idx] = similarity_data(r1, r2)
    res = 1 - 2 * np.average(np.absolute(.5 - similarities), axis=0)
    return res


def uniqueness(instances: List[Simulation], seed: int, N: int = 10000) -> np.ndarray:
    r"""
    Estimates the uniqueness of a list of given Simulations.

    Randomly generates :math:`N` challenges using seed :math:`\texttt{seed}`, then queries each simulation
    each challenge, obtaining an array with shape :math:`(l, N, m)` of all responses, where
    :math:`l` is the number of simulations, and :math:`m` is the response length of the given simulations. (All
    given simulations must have same challenge and response lenght.)

    Then applies :func:`uniqueness_data` to determine the uniqueness of each response bit.

    Returns a float-array of shape :math:`(m,)`, giving the uniqueness of each output bit.
    To obtain total uniqueness, average along all axes, but be aware that low uniqueness on individual response bits
    may be problematic.

    >>> from numpy import average
    >>> from pypuf.simulation import XORArbiterPUF
    >>> from pypuf.metrics import uniqueness
    >>> instances = [XORArbiterPUF(n=64, k=1, seed=seed) for seed in range(5)]
    >>> uniqueness(instances, seed=31415, N=1000)
    array([0.9272])
    """
    m = instances[0].response_length
    challenges = random_inputs(instances[0].challenge_length, N, seed)
    responses = np.empty(shape=(len(instances), N, m))
    for i, instance in enumerate(instances):
        responses[i] = instance.eval(challenges).reshape(N, m)  # TODO fixes incompatible LTFArray output format
    return uniqueness_data(responses)


def similarity_data(responses1: np.ndarray, responses2: np.ndarray) -> np.ndarray:
    r"""
    Given two arrays of responses of shape :math:`(N, m)`, returns the relative frequency of equal responses
    in the arrays for each response bit, i.e. an approximation of

    .. math:: \Pr_x \left[ f(x) = g(x) \right],

    where :math:`f` and :math:`g` are the functions given by :math:`\mathtt{responses1}` and
    :math:`\mathtt{responses2}`, respectively; :math:`x` are the challenges (given implicitly by ordering of the
    response arrays).

    If response length :math:`m` is greater than 1, the similarity is given for each bit seperately. To obtain the
    general similarity, average the response.

    Returns an array of shape :math:`(m,)`.

    >>> from pypuf.metrics import similarity_data
    >>> from numpy import array
    >>> similarity_data(array([[1, 1], [1, 1], [1, 1], [1, 1]]), array([[1, 1], [1, 1], [1, 1], [-1, 1]]))
    array([0.75, 1.  ])
    """
    # Check for response arrays with repetitions (shape (N, m, r))
    # and majority vote along the repetition axis
    if len(responses1.shape) == 3:
        responses1 = np.sign(np.average(responses1, axis=-1))
    if len(responses2.shape) == 3:
        responses2 = np.sign(np.average(responses2, axis=-1))

    # check for broken response arrays with implicit response length 1 (shape (N,))
    # TODO fix LTFArray to return responses of shape (N, 1)
    if len(responses1.shape) == 1:
        responses1 = responses1.reshape(-1, 1)
    if len(responses2.shape) == 1:
        responses2 = responses2.reshape(-1, 1)

    return 1 - np.average(np.absolute(responses1 - responses2), axis=0) / 2


def accuracy(simulation: Simulation, test_set: ChallengeResponseSet) -> np.ndarray:
    r"""
    Given a simulation and a test set, returns the relative frequency of responses by the simulation that match
    responses recorded in the test set by response bit, i.e. an approximation of

    .. math:: \Pr_x \left[ f(x) = g(x) \right],

    where :math:`f` is given by the provided simulation and :math:`g` is defined by the :math:`\mathtt{test\_set}`;
    :math:`x` is from the set of challenges given in the :math:`\mathtt{test\_set}`.

    If response length is greater than 1, the similarity is given for each bit seperately. To obtain the
    general similarity, average the response.

    Returns an array of shape :math:`(m,)`.

    >>> from pypuf.simulation import XORArbiterPUF
    >>> from pypuf.io import ChallengeResponseSet
    >>> from pypuf.metrics import accuracy
    >>> puf = XORArbiterPUF(n=128, k=4, noisiness=.1, seed=1)
    >>> test_set = ChallengeResponseSet.from_simulation(puf, N=1000, seed=2)
    >>> accuracy(puf, test_set)
    array([0.823])
    >>> puf = XORArbiterPUF(n=64, k=4, noisiness=.3, seed=2)
    >>> test_set = ChallengeResponseSet.from_simulation(puf, N=1000, seed=2, r=5)
    >>> accuracy(puf, test_set)
    array([0.706])
    """
    sim_responses = simulation.eval(test_set.challenges)
    return similarity_data(sim_responses, test_set.responses)


def similarity(instance1: Simulation, instance2: Simulation, seed: int, N: int = 1000) -> np.ndarray:
    r"""
    Approximate the similarity in response behavior of two simulations with identical challenge and response length,
    i.e. an approximation of

    .. math:: \Pr_x \left[ f(x) = g(x) \right],

    where :math:`f` and :math:`g` are given by the provided simulations, and the :math:`N` challenges :math:`x` are generated
    randomly using the provided :math:`\mathtt{seed}`.

    If response length is greater than 1, the similarity is given for each bit seperately. To obtain the
    general similarity, average the response.

    Returns an array of shape :math:`(m,)`.

    >>> from pypuf.simulation import XORArbiterPUF
    >>> from pypuf.metrics import similarity
    >>> similarity(XORArbiterPUF(n=128, k=4, seed=1), XORArbiterPUF(n=128, k=4, seed=1), seed=31415)  # same seed
    array([1.])
    >>> similarity(XORArbiterPUF(n=128, k=4, seed=1), XORArbiterPUF(n=128, k=4, seed=2), seed=31415)  # different seed
    array([0.515])
    """
    if instance1.challenge_length != instance2.challenge_length:
        raise ValueError(f'Cannot compare instances with different challenge spaces of dimension '
                         f'{instance1.challenge_length} and {instance2.challenge_length}, respectively.')
    inputs = random_inputs(n=instance1.challenge_length, N=N, seed=seed)
    return similarity_data(instance1.eval(inputs), instance2.eval(inputs))



import logging
import os
import pickle
import random
from datetime import datetime, timedelta
from hashlib import sha256
from pathlib import Path
from typing import List, Optional

import pandas as pd
from numpy.distutils import cpuinfo
from memory_profiler import memory_usage

# noinspection PyBroadException
try:
    cpu = cpuinfo.cpu.info[0]['model name']
except Exception:
    cpu = None


class ResultCollection:

    def __init__(self) -> None:
        pass

    def add_result(self, parameter_hash: str, result: dict) -> None:
        raise NotImplementedError

    def known_results(self) -> List[str]:
        raise NotImplementedError

    def save_log(self, log: object, params: dict, parameter_hash: str, force: bool = False) -> None:
        raise NotImplementedError


class PickleResultCollection(ResultCollection):

    def __init__(self, path: str) -> None:
        super().__init__()
        self.path = path
        self.log_path = self.path + '.log.pickle'
        self.log_saved = datetime.now()
        try:
            self.results = pd.read_pickle(self.path)
        except FileNotFoundError:
            self.results = pd.DataFrame()

    def add_result(self, parameter_hash: str, result: dict) -> None:
        self.results = self.results.append(result, ignore_index=True)
        self._save_results()

    def _save_results(self) -> None:
        if not self.path:
            return
        logging.debug(f'saving results file {self.path} ({len(self.results)} results)')
        self.results.to_pickle(f'{self.path}.pickle')
        self.results.to_csv(f'{self.path}.csv')

    def known_results(self) -> List[str]:
        return list(self.results.get('param_hash', []))

    def save_log(self, log: object, params: dict, parameter_hash: str, force: bool = False) -> None:
        if not self.log_path:
            return
        if force or datetime.now() - self.log_saved > timedelta(minutes=10):
            with open(self.log_path, 'wb') as f:
                pickle.dump((params, log), f)
                self.log_saved = datetime.now()


class FilesystemResultCollection(ResultCollection):

    EXTENSION = '.pickle'
    LOG_EXTENSION = '.log'

    def __init__(self, path: str) -> None:
        super().__init__()
        self.path = path
        if not Path(path).exists():
            Path(path).mkdir()
        else:
            if not Path(path).is_dir():
                raise ValueError(f'Given storage path {Path(path).absolute()} must be '
                                 f'either non-existent or a directory.')

        self.log_saved = datetime.now()  # TODO Consider splitting throttling by parameter hash

    def add_result(self, parameter_hash: str, result: dict) -> None:
        with open(Path(self.path) / (parameter_hash + self.EXTENSION), 'wb') as f:
            pickle.dump(result, f)

    def _result_pickles(self) -> List[str]:
        return list(filter(lambda x: x.endswith(self.EXTENSION), os.listdir(self.path)))

    def known_results(self) -> List[str]:
        pickles = self._result_pickles()
        return [p.split('.')[0] for p in pickles]

    def save_log(self, log: object, params: dict, parameter_hash: str, force: bool = False) -> None:
        if force or datetime.now() - self.log_saved > timedelta(minutes=10):
            with open(Path(self.path) / (parameter_hash + self.EXTENSION + self.LOG_EXTENSION), 'wb') as f:
                pickle.dump((params, log), f)
                self.log_saved = datetime.now()

    def load_all(self) -> List[dict]:
        pickles = self._result_pickles()
        results = []
        for p in pickles:
            with open(Path(self.path) / p, 'rb') as f:
                results.append(pickle.load(f))
        return results


class MemoryResultCollection(ResultCollection):

    def __init__(self) -> None:
        super().__init__()
        self.results = {}
        self.log = {}

    def add_result(self, parameter_hash: str, result: dict) -> None:
        self.results[parameter_hash] = result

    def known_results(self) -> List[str]:
        return list(map(str, self.results.keys()))

    def save_log(self, log: object, params: dict, parameter_hash: str, force: bool = False) -> None:
        self.log[parameter_hash] = (params, log)

    def load_all(self) -> List[dict]:
        return list(self.results.values())


class StudyBase:

    def __init__(self, results: Optional[ResultCollection] = None, logging_callback: Optional[callable] = None,
                 randomize_order: bool = True) -> None:
        self._timer = {}

        if isinstance(results, str):
            results = FilesystemResultCollection(results)
        elif results is None:
            results = MemoryResultCollection()
        self.results = results

        self.log = None
        self.logging_callback = logging_callback

        self._cached_parameter_matrix = self.parameter_matrix()
        self._randomized_order = randomize_order
        if randomize_order:
            random.Random(42).shuffle(self._cached_parameter_matrix)
        self._current_params = None

        self.continue_on_error = False

    def _start_timer(self, name: str = 'default') -> None:
        self._timer[name] = datetime.now()

    def _stop_timer(self, name: str = 'default') -> float:
        return datetime.now() - self._timer[name]

    def _hash_parameters(self, params: dict) -> str:
        return sha256((self.__class__.__name__ + ': ' + str(params)).encode()).hexdigest()

    @staticmethod
    def parameter_matrix() -> List[dict]:
        raise NotImplementedError

    def primary_results(self, results: dict) -> dict:
        raise NotImplementedError

    def _add_result(self, params: dict, result: dict, memory: list = None) -> None:
        row = {}
        row.update({
            'parameters': list(map(str, params.keys())),
            'results': list(map(str, result.keys())),
            'experiment': self.__class__.__name__,
            'cpu': cpu,
            'timestamp': datetime.now(),
            'param_hash': self._hash_parameters(params),
            'memory': memory,
            # TODO consider adding git state
        })
        for env_var in [
            'OMP_NUM_THREADS',
            'NUMEXPR_NUM_THREADS',
            'MKL_NUM_THREADS',
            'TF_NUM_INTRAOP_THREADS',
            'TF_NUM_INTEROP_THREADS',
        ]:
            row.update({env_var: os.environ.get(env_var, None)})
        row.update(params)
        row.update(result)
        self.results.add_result(row['param_hash'], row)
        try:
            logging.debug(f'Added result: {self.primary_results(row)}')
        except NotImplementedError:
            logging.debug('Added result.')

    def _save_log(self, force: bool = False) -> None:
        if callable(self.logging_callback):
            self.logging_callback()
        self.results.save_log(self.log, self._current_params, self._hash_parameters(self._current_params), force)

    def run(self, **kwargs: dict) -> dict:
        raise NotImplementedError

    def run_single(self, params: dict) -> None:
        logging.debug(f'Running {self.__class__.__name__} for {params}')

        memory, result = memory_usage((self.run, [], params), retval=True)
        self._add_result(params, result, memory)

    def run_block(self, index: int, total: int) -> None:
        parameter_matrix = self._cached_parameter_matrix
        n = len(parameter_matrix)
        self.run_batch(parameter_matrix[int(index / total * n):int((index + 1) / total * n)])

    def run_all(self) -> None:
        self.run_batch(self._cached_parameter_matrix)

    def run_batch(self, batch: list) -> None:
        unfinished_parameters = [
            params for params in batch if self._hash_parameters(params) not in self.results.known_results()
        ]

        total = len(self._cached_parameter_matrix)
        unfinished = len(unfinished_parameters)
        finished = len(batch) - unfinished

        logging.debug(f'{self.__class__.__name__}: running {unfinished} unfinished jobs from a batch of {len(batch)} '
                      f'({finished} of this batch already completed, batch total {len(batch)}, '
                      f'study total {total} jobs)')

        ran = 0
        for params in unfinished_parameters:
            logging.debug(f'Progress: {ran/unfinished:.1%} session, {(ran + finished)/len(batch):.1%} batch, '
                          f'{(ran + finished)/total:.1%} total')
            try:
                ran += 1
                self._current_params = params
                self.run_single(params)
            except Exception as e:
                logging.debug(f'Running {self.__class__.__name__} resulted in {type(e)}: {e}.')
                if not self.continue_on_error:
                    raise
            finally:
                self._current_params = None

    @classmethod
    def cli(cls, args: list) -> None:
        logging.basicConfig(format='%(asctime)s.%(msecs)06d: %(levelname)s %(message)s',
                            level=logging.DEBUG,
                            datefmt='%Y-%m-%d %H:%M:%S')

        randomize = len(args) > 4
        study = cls(args[1], randomize_order=randomize)
        block_idx = int(args[2])
        block_total = int(args[3])
        study.run_block(block_idx, block_total)
